//////////////////////////////////////////////////////////////////////////////
//
// File: expr.cpp
//
// expr.cpp contains the implementation of the entire tree-structure for the
// expression trees. All classes are declared in expr.h, which is the only 
// file needed for using expression trees.
//
// The expression trees are completely based on pointers, this means that you
// need to be extra careful when you do and don't delete the pointers. 
// Basically - when creating trees with the constructors, you should never
// delete the expressions you used to build new expression. When you subst(),
// diff() or simplify() you get an entire new tree you may use as you find
// appropriate - the original tree is not changed in any way.
//
// Two nice macroes are defined - DEBUG to keep track of debugging info and
// DESTRUCTOR to monitor destruction of objects.
#include <math.h>
#include "expr.h"

//#define DEBUG(s) (cerr << s << endl)
#define DEBUG(s)
//#define DESTRUCTOR(s) (cerr << s << endl)
#define DESTRUCTOR(s)



///////////////////////////////////////////////////////////////////////////////
//
// Different general functions
//   - overloading << for Xpr
//

// Overloading the << operator for all Xpr objects
ostream& operator<<(ostream& o, Xpr& e) {
	return ( o << e.toString() );
};

// Special function for indenting
void indenter(int x) {
	for (;x>0;x--) cout << '\t';
}




///////////////////////////////////////////////////////////////////////////////
//
//  X c o n s t b o o l - definition of constant for bool
//
///////////////////////////////////////////////////////////////////////////////


// Constructor
Xconstbool::~Xconstbool() {
	DESTRUCTOR("~Xconstbool: " << this );
}


// Raise exception
pXpr Xconstbool::diff( Xvar& x ) const {
	throw Exception(E_DIFF_BOOL);
}


// Return value
Value Xconstbool::evalf( Enviroment& env ) const {
	Value e;
	e.type = Value::BOOL;
	e.val.bval = value;
	return e;
}


// Constant cannot be simplified
pXpr Xconstbool::simplify() const {
	return clone();
}


// Clone method
pXpr Xconstbool::clone() const {
	return new Xconstbool(value);
}


// toString method
string Xconstbool::toString(int=0) const {
	if (value) return "true";
	else return "false";
}


// Support for comparing - type and value must match
bool Xconstbool::isEqual(pXpr expr) const {
	Xconstbool* p = dynamic_cast<Xconstbool*>(expr);
	if (!p) return false;
	else return (p->value == value);
}


// Substituting expressions - nothing to do here.
pXpr Xconstbool::subst( Xvar&, pXpr ) const {
	return clone();
}




///////////////////////////////////////////////////////////////////////////////
//
// X c o n s t r e a l - definition of constant containing real
//
///////////////////////////////////////////////////////////////////////////////


// Constructor
Xconstreal::~Xconstreal() {
	DESTRUCTOR("~Xconstreal: " << this );
}


// Constant diff'd to 0.0
pXpr Xconstreal::diff (Xvar &x) const {
	return new Xconstreal(0.0); 
}


// Return value
Value Xconstreal::evalf( Enviroment& env ) const {
	Value e;
	e.type = Value::REAL;
	e.val.fval = value;
	return e;
}


// If return itself
pXpr Xconstreal::simplify() const {
	return clone();
}


// Clone method
pXpr Xconstreal::clone() const {
	return new Xconstreal(value);
}


// toString method
string Xconstreal::toString(int=0) const {
	//return "<const>";
	ostrstream s;
	s << value << ends;
	return string(s.str());
}


// Compare - type and value must match
bool Xconstreal::isEqual(pXpr expr) const {
	Xconstreal* p = dynamic_cast<Xconstreal*>(expr);
	if (p) return (p->value == value);
	else return false;
}


// Substituting expressions - nothing to do here.
pXpr Xconstreal::subst( Xvar&, pXpr ) const {
	return clone();
}




///////////////////////////////////////////////////////////////////////////////
//
//  X v a r - atom in expression
//
///////////////////////////////////////////////////////////////////////////////


// Constructor - protects against creating a Xvar without a name
Xvar::Xvar( string nname, exprType et=BOTH ): name(nname), vartype(et) {
	assert( name != "" );
}


// Copy-constructor
Xvar::Xvar( const Xvar& x ): name(x.name), vartype(x.vartype) {
}


// Destroy Xvar
Xvar::~Xvar() {
	DESTRUCTOR("~Xvar: " << this );
}

string Xvar::toString(int=0) const {
/*	string tp;
	switch (vartype) {
		case NONE: tp = "NONE"; break;
		case BOTH: tp = "BOTH"; break;
		case REAL: tp = "REAL"; break;
		case BOOL: tp = "BOOL"; break;
	}
	return name + "{" + tp + "}"; */
	return name;
};


// Differentiate current var with respect to Xvar supplied - if match then diff'd to 1.0 else 0.0
pXpr Xvar::diff( Xvar& x ) const {
	DEBUG("Xvar::diff(): attempt to diff '" << name << " with regard to: " <<  x );
	if (x.name == name)
		return new Xconstreal(1.0);
	else
		return new Xconstreal(0.0);
}


// Evaluate this variable with respect to the enviroment. If the enviroment does not
// contain the identifier, cause an exception since none is returned
Value Xvar::evalf( Enviroment& env ) const {
	Value e;
	Xconst* x = dynamic_cast<Xconst*>(env[*this]);
	if (x) {
		if (dynamic_cast<Xconstbool*>(x)) {
			e.val.bval = dynamic_cast<Xconstbool*>(x)->getVal();
			e.type = Value::BOOL;
		}
		if (dynamic_cast<Xconstreal*>(x)) {
			e.val.fval = dynamic_cast<Xconstreal*>(x)->getVal();
			e.type = Value::REAL;
		}
	} else 
		throw Exception(E_EVAL_NOTCONST, name);
	return e;
}

		
// simplify() returns a new Xvar with same contents.
pXpr Xvar::simplify() const {
	DEBUG("Xvar::simplify(): simplifying var " << name );
	return clone();
}


// Clone this object
pXpr Xvar::clone() const {
	return new Xvar(*this);
}


// Compare - type and name must match
bool Xvar::isEqual(pXpr expr) const {
	Xvar* p = dynamic_cast<Xvar*>(expr);
	if (p) return (p->name == name);
	else return false;
}


// Substituting expressions - if variable name matches, then return clone of pXpr else return clone of self
pXpr Xvar::subst( Xvar& x, pXpr p) const {
	if (x.name==name) return p->clone();
	else return clone();
}


// The map<>-template requires us to implement a method to compare two Xvar's
bool Xvar::operator<(const Xvar& x) const {
	return (x.name<name);
}




///////////////////////////////////////////////////////////////////////////////
//
//  X u n a r y - definition of functions for unary operator
// 
///////////////////////////////////////////////////////////////////////////////

// Normal constructor
Xunary::Xunary( string fun, pXpr args ): Xpr(), fname(fun), arg(args) {
	assert( fname!="" && arg );
	     if (fname=="sin") func = SIN;
	else if (fname=="cos") func = COS;
	else if (fname=="ln")  func = LN;
	else if (fname=="exp") func = EXP;
	else if (fname=="tan") func = TAN;
	else if (fname=="not") func = NOT;
	else if (fname=="!") func = NOT;
	else if (fname=="sinh") func = SINH;
	else if (fname=="cosh") func = COSH;
	else func = OTHER;
	DEBUG("Xunary::Xunary() - " << fname << "  " << func);
}


// Deep-copying constructor
Xunary::Xunary( const Xunary& b): Xpr(), fname(b.fname), func(b.func) {
	arg = (b.arg)->clone();
}


// Destructor - delete argument-tree
Xunary::~Xunary() { 
	DESTRUCTOR("~Xunary: " << this);
	delete arg; 
}


// Differentiates according to simple rules. If it knows the function (like ln/exp/sin/cos) it differentiates accordingly, otherwise
// it differentiates by applying ' ( like f(2*x) = 2*f'(2*x) )
pXpr Xunary::diff( Xvar& x ) const {
	DEBUG("Xunary::diff():");
	pXpr inside = NULL;
	try {
	  inside = arg->diff(x);
	  switch (func) {
	  	case SIN: // sin'(f) = cos(f)*f'
	  		return new Xbinary( new Xunary( "cos", arg->clone() ), Xbinary::MUL, inside );
	  		
	  	case COS: // cos'(f) = 0-sin(f)*f'
	  		return new Xbinary( new Xconstreal(0.0), Xbinary::SUB, new Xbinary( new Xunary( "sin", arg->clone() ), Xbinary::MUL, inside ) );
	  		
	  	case EXP: // exp'(f) = exp(f)*f'
	  		return new Xbinary( new Xunary( fname, arg->clone() ), Xbinary::MUL, inside );
	  		
	  	case LN: // ln'(f) = f'/f
	  		return new Xbinary( inside, Xbinary::DIV, arg->clone() );
	  		
	  	case TAN: // tan'(f) = f'*(1+tan(f)^2) = f'/(cos(f)^2)
	  		return new Xbinary( inside, Xbinary::MUL, new Xbinary(new Xconstreal(1.0), Xbinary::ADD, new Xbinary(clone(), Xbinary::MUL, clone())));
	  		//return new Xbinary( inside, Xbinary::DIV, new Xbinary(new Xunary("cos",arg->clone()), Xbinary::MUL, new Xunary("cos",arg->clone())) );
	  		
	  	case SINH:
	  		return new Xbinary( inside, Xbinary::MUL, new Xunary("cosh", arg->clone()));

	  	case COSH:
	  		return new Xbinary( inside, Xbinary::MUL, new Xunary("sinh", arg->clone()));
	  		
	  	case NOT: // not is boolean function and can thus not be differentiated.
	  		throw Exception(E_DIFF_BOOL, toString() );
	  		
	  	default: 
	  		return new Xbinary(new Xunary( fname+"'", arg->clone() ), Xbinary::MUL, inside );
	  }
	} catch (Exception e) {
	  if (inside) delete inside;
	  throw;
	}
}


// Attempts to evaluate expression. Known functions are calculated, NONE returned otherwise
Value Xunary::evalf( Enviroment& env ) const {
	Value e;
	e = arg->evalf(env);
	e = calculate(e);
	return e;
}


// Calculates the functional value of a function based on a previously calculated value for the argument
// If there is a type-error, an exception is thrown.
Value Xunary::calculate( Value v ) const {
	switch (func) {
		case SIN: 
			if (v.type == Value::REAL) v.val.fval = sin(v.val.fval);
			else throw Exception(E_EVAL_BOOL, toString() );
			break;
		
		case COS:
			if (v.type == Value::REAL) v.val.fval = cos(v.val.fval);
			else throw Exception(E_EVAL_BOOL, toString() );
			break;

		case SINH: 
			if (v.type == Value::REAL) v.val.fval = sinh(v.val.fval);
			else throw Exception(E_EVAL_BOOL, toString() );
			break;

		case COSH: 
			if (v.type == Value::REAL) v.val.fval = cosh(v.val.fval);
			else throw Exception(E_EVAL_BOOL, toString() );
			break;

		
		case EXP:
			if (v.type == Value::REAL) v.val.fval = exp(v.val.fval);
			else throw Exception(E_EVAL_BOOL, toString() );
			break;
			
		case LN:
			if (v.type == Value::REAL) {
				if (v.val.fval>0) v.val.fval = log(v.val.fval);
				else throw Exception(E_EVAL_NEGLN);				
			} else throw Exception(E_EVAL_BOOL, toString() );
			break;

		case TAN:
			if (v.type == Value::REAL) v.val.fval = tan(v.val.fval);
			else throw Exception(E_EVAL_BOOL, toString() );
			break;

		case NOT:
			if (v.type == Value::BOOL) v.val.bval = !(v.val.bval);
			else throw Exception(E_EVAL_REAL, toString() );
			break;
			
		default:
			throw Exception(E_EVAL_NORULE, toString());
	}
	return v;
}


// If simplify(subexpr) is Xconst and function is known then return calculated value as Xconst
pXpr Xunary::simplify() const {
	DEBUG("Xunary::simplify");
	pXpr simp = NULL;
	try { 
		simp = arg->simplify();
		// If known function and simplify(arg) is constant, then apply function. If type-error, exception is thrown
		if (func!=OTHER && dynamic_cast<Xconst*>(simp)) {
			DEBUG(" - simplifying func(constant)");
			Value v = dynamic_cast<Xconst*>(simp)->evalf();
			// calculate value 
			v = calculate(v);
			if (v.type==Value::REAL) {
				delete simp;
				return new Xconstreal(v.val.fval);
			}
			if (v.type==Value::BOOL) {
				delete simp;
				return new Xconstbool(v.val.bval);
			}
			return new Xunary( fname, simp );
		} else // !(!(x)) = x
			if (func==NOT && dynamic_cast<Xunary*>(simp) && dynamic_cast<Xunary*>(simp)->func==NOT) return dynamic_cast<Xunary*>(simp)->arg->clone();
		else // exp(ln(x)) = x  (unsupported x>0)
			if (func==EXP && dynamic_cast<Xunary*>(simp) && dynamic_cast<Xunary*>(simp)->func==LN) return dynamic_cast<Xunary*>(simp)->arg->clone();
		else // ln(exp(x)) = x  (unsupported x>0)
			if (func==LN && dynamic_cast<Xunary*>(simp) && dynamic_cast<Xunary*>(simp)->func==EXP) return dynamic_cast<Xunary*>(simp)->arg->clone();
		else
			return new Xunary( fname, simp );
	} catch (Exception e) {
		if (simp) delete simp;
		throw;
	}
}


// Return string representation of object
// NOT is considered a unary operator not needing () around its argument if it is a very simple one (Xvar)
// thus it sets its parent-priority to 100
string Xunary::toString(int pp=0) const {
  if (func==NOT) return fname+arg->toString(100);
  else return fname +"("+ arg->toString() +")";
}


// Clone method - uses deep-copying constructor
pXpr Xunary::clone() const {
	return new Xunary(*this);
}


// Support for comparing - type, name and argument must match
bool Xunary::isEqual(pXpr expr) const {
	Xunary* p = dynamic_cast<Xunary*>(expr);
	if (p) return (p->fname==fname && arg->isEqual(p->arg) );
	else return false;
}


// Substituting expressions - substitute in argument and keep function
pXpr Xunary::subst( Xvar& x, pXpr p) const {
	return new Xunary( fname, arg->subst(x,p) );
}

// Checks types - not is the only bool->bool function - all others are assumed to be real->real!
Xpr::exprType Xunary::getType() const {
	exprType at = arg->getType();
	if (func==NOT) return (at==BOOL?BOOL:NONE);
	return (at==REAL?REAL:NONE); 
};


///////////////////////////////////////////////////////////////////////////////
//
//  X b i n a r y - binary operators
//
///////////////////////////////////////////////////////////////////////////////


// Implement binary operators like '+' , '-' , '*' , '/'
Xbinary::Xbinary( pXpr l, opType o, pXpr r): left(l), right(r), op(o) {
	assert( left && right );
}


// Deep-copying constructor
Xbinary::Xbinary( const Xbinary& b ): op(b.op) {
	left = (b.left)->clone();
	right = (b.right)->clone();
	assert( left && right );
}


// Destructor - delete both sub-trees
Xbinary::~Xbinary() { 
	DESTRUCTOR("~Xbinary: " << this );
	if (left) delete left;
	if (right) delete right; 
};


// Differentiates according to simple rules. Throws exception if opType is OR/AND/LT/LTEQ/GT/GTEQ/NOTEQ
pXpr Xbinary::diff( Xvar& x ) const {
	pXpr leftdiff = NULL;
	pXpr rightdiff = NULL;
	pXpr leftnew = NULL;
	pXpr rightnew = NULL;
	try {
		switch (op) {
		case OR: case AND: case LT: case LTEQ: case GT: case GTEQ: case EQ: case NOTEQ:
			throw Exception(E_DIFF_BOOL);
			
		case ADD: case SUB:
			leftdiff = left->diff(x);
			rightdiff = right->diff(x);
			return new Xbinary( leftdiff, op, rightdiff );
			
		case MUL: 
			leftdiff = left->diff(x);
			rightdiff = right->diff(x);
			leftnew = left->clone();
			rightnew = right->clone();
			return new Xbinary( new Xbinary(leftdiff,MUL,rightnew), ADD, new Xbinary(leftnew,MUL,rightdiff) );	  
			
		case DIV: 
			leftdiff = left->diff(x);
			rightdiff = right->diff(x);
			leftnew = new Xbinary( new Xbinary(leftdiff,MUL,right->clone()), SUB, new Xbinary(left->clone(),MUL,rightdiff) );
			rightnew = new Xbinary( right->clone(), MUL, right->clone() );
			return new Xbinary( leftnew, DIV, rightnew );	  
	
		case POW: {
			bool expindep = false;
			bool otindep = false;
			
			rightdiff = right->subst(x,new Xconstreal(0.0));
			expindep = *right==*rightdiff;
			delete rightdiff; rightdiff=NULL;
			leftdiff = left->subst(x,new Xconstreal(0.0));
			otindep = *left==*leftdiff;
			delete leftdiff; leftdiff=NULL;
			if (expindep && !otindep) { // exponent independent of x
				rightdiff = new Xbinary(right->clone(), SUB, new Xconstreal(1.0));
				rightnew = new Xbinary(left->clone(), POW, rightdiff);
				return new Xbinary(right->clone(), MUL, rightnew);
			} 
			if (otindep && !expindep) {
	  			rightdiff = right->diff(x);
	  			rightnew = new Xbinary(rightdiff, MUL, new Xunary("ln", left->clone()) );
	  			return new Xbinary( rightnew, MUL, clone() );
			}
			throw Exception(E_DIFF_POWNOTINDEP, toString());
		}
		} //switch
		  
	} catch (Exception e) {
	  if (leftnew) delete leftnew;
	  if (rightnew) delete rightnew;
	  if (leftdiff) delete leftdiff;
	  if (rightdiff) delete rightdiff;
	  throw;
	}
}


// Attempts to evaluate expression
Value Xbinary::evalf( Enviroment& env ) const {
	Value lv, rv;
	lv = left->evalf(env);
	rv = right->evalf(env);
	return calculate(lv,rv);
}


// Helper method to perform the actual calculations
Value Xbinary::calculate(Value lv, Value rv) const {
	Value res;
	res.type = Value::NONE;
	
	// Check for type-errors
	switch (op) {
		case OR: case AND:
			if (lv.type!=Value::BOOL || rv.type!=Value::BOOL) 
				throw Exception(E_EVAL_REAL,toString());
			break;
			
		default:
			if (lv.type!=Value::REAL || rv.type!=Value::REAL) 
				throw Exception(E_EVAL_BOOL,toString());
			break;
	}
	
	// now we can assume, that no type-errors exist in the values, so just evaluate
	switch (op) {
		case OR:
			res.type = Value::BOOL;
			res.val.bval = lv.val.bval || rv.val.bval;
			break;
			
		case AND:
			res.type = Value::BOOL;
			res.val.bval = lv.val.bval && rv.val.bval;
			break;

		case LT:
			res.type = Value::BOOL;
			res.val.bval = lv.val.fval < rv.val.fval;
			break;

		case LTEQ:
			res.type = Value::BOOL;
			res.val.bval = lv.val.fval <= rv.val.fval;
			break;

		case GT:
			res.type = Value::BOOL;
			res.val.bval = lv.val.fval > rv.val.fval;
			break;

		case GTEQ:
			res.type = Value::BOOL;
			res.val.bval = lv.val.fval >= rv.val.fval;
			break;

		case EQ:
			res.type = Value::BOOL;
			res.val.bval = lv.val.fval == rv.val.fval;
			break;

		case NOTEQ:
			res.type = Value::BOOL;
			res.val.bval = lv.val.fval != rv.val.fval;
			break;

		case ADD:
			res.type = Value::REAL;
			res.val.fval = lv.val.fval + rv.val.fval;
			break;

		case SUB:
			res.type = Value::REAL;
			res.val.fval = lv.val.fval - rv.val.fval;
			break;

		case MUL:
			res.type = Value::REAL;
			res.val.fval = lv.val.fval * rv.val.fval;
			break;

		case DIV:
			res.type = Value::REAL;
			if (rv.val.fval==0.0) throw Exception(E_DIV_BY_ZERO);
			res.val.fval = lv.val.fval / rv.val.fval;
			break;
						
		case POW: // currently no domain-check at all!
			res.type = Value::REAL;
			//if (lv.val.fval<=0.0) throw Exception(E_EVAL_POWNEGBASE);
			res.val.fval = pow(lv.val.fval,rv.val.fval);
			break;
			
		default:
			res.type = Value::NONE;
	}	
	return res;
}



// get type for this expression. If type-error exists, NONE is returned
Xpr::exprType Xbinary::getType() const {
	switch (op) {
		case OR: case AND: 
			if ( (left->getType()==BOOL || left->getType()==BOTH) && (right->getType()==BOOL || right->getType()==BOTH) )
				return BOOL;
			else
				return NONE;
			
		case LT: case LTEQ: case GT: case GTEQ: case EQ: case NOTEQ:
			if ( (left->getType()==REAL || left->getType()==BOTH) && (right->getType()==REAL || right->getType()==BOTH) )
				return BOOL;
			else
				return NONE;
		
		
		case ADD: case SUB: case MUL: case DIV: case POW:
			if ( (left->getType()==REAL || left->getType()==BOTH) && (right->getType()==REAL || right->getType()==BOTH) )
				return REAL;
			else
				return NONE;
			
		default:
			return NONE;
	}
}


// simplifies sub-expressions and then "Xconst op Xconst"
//
// Rules supported:
//  - <const> op <const> simplified
//  - <const=0,1> * <expr> simplified to <expr>
//  - <expr> *,/ <const=0,1> simplified to <expr> or exception
//  ! <bool> AND <expr> , <bool> OR <expr> (and inverse)
pXpr Xbinary::simplify() const {
	DEBUG("Xbinary::simplify():");
	pXpr leftsimp = NULL; 
	pXpr rightsimp = NULL;
	try {
		leftsimp = left->simplify();
		rightsimp = right->simplify();
		Xconst* lc = dynamic_cast<Xconst*>(leftsimp);
		Xconst* rc = dynamic_cast<Xconst*>(rightsimp);

	
		if (lc && rc) { // <const> op <const> => <const>
			DEBUG("\t<const> op <const>");
			Value lv, rv;
			// convert the two constants to values and calculate (may throw exceptions)
			lv = lc->evalf();
			rv = rc->evalf();
			lv = calculate(lv,rv);
			delete leftsimp; 
			delete rightsimp; 
			
			switch (lv.type) {
				case Value::BOOL: return new Xconstbool(lv.val.bval);
				case Value::REAL: return new Xconstreal(lv.val.fval);
			}
		
		} else 
		if (lc) { // <const> op <exp>
			pXpr res = simp_const_op_expr(lc,rightsimp);
			if (res) return res;
			
		} else
		// <exp> op <const>
		if (rc) {
			pXpr res = simp_expr_op_const(leftsimp,rc);
			if (res) return res;			
		} else
		// operator= and <leftexp> = <rightexp>
		if (op==EQ && *leftsimp==*rightsimp) {
			delete leftsimp;
			delete rightsimp;
			return new Xconstbool(true);
		}
		
		DEBUG("\tnothing in this level");
  		return new Xbinary( leftsimp, op, rightsimp );
	} catch (Exception e) {
		if (leftsimp) delete leftsimp;
		if (rightsimp) delete rightsimp;
		throw;
	}
}


pXpr Xbinary::simp_const_op_expr(Xconst* lc, Xpr* r) const {
	DEBUG("\t<const>*<exp>");
	Xconstbool* lb = dynamic_cast<Xconstbool*>(lc);
	Xconstreal* lr = dynamic_cast<Xconstreal*>(lc);
	switch (op) {
		case OR: // true OR ? => true
			if (!lb) 
				throw Exception(E_INVALID_SUBTYPE);
			if (lb->getVal()) {
				delete lc; delete r;
				return new Xconstbool(true);
			}
			break;
			
		case AND:
			if (!lb) // false AND ? => false
				throw Exception(E_INVALID_SUBTYPE);
			if ( !(lb->getVal()) ) {
				delete lc; delete r;
				return new Xconstbool(false);
			}
			break;
				
		case MUL:
			if (!lr) 
				throw Exception(E_INVALID_SUBTYPE);
			if ( lr->getVal()==0.0) {
				delete lc; delete r;
				return new Xconstreal(0.0);
			}
			if ( lr->getVal()==1.0) {
				delete lc;
				return r;
			}
			break;

		case DIV:
			if (!lr) 
				throw Exception(E_INVALID_SUBTYPE);
			if ( lr->getVal()==0.0) {
				delete lc; delete r;
				return new Xconstreal(0.0);
			}
			break;
			
		case ADD:
			if (!lr)
				throw Exception(E_INVALID_SUBTYPE);
			if ( lr->getVal()==0.0) {
				delete lc; 
				return r;
			}
	}
	return NULL;	
}

pXpr Xbinary::simp_expr_op_const(Xpr* l, Xconst *rc) const {
	DEBUG("\t<const>*<exp>");
	Xconstbool* rb = dynamic_cast<Xconstbool*>(rc);
	Xconstreal* rr = dynamic_cast<Xconstreal*>(rc);
	switch (op) {
		case OR: // true OR ? => true
			if (!rb) 
				throw Exception(E_INVALID_SUBTYPE);
			if (rb->getVal()) {
				delete rc; delete l;
				return new Xconstbool(true);
			}
			break;
			
		case AND:
			if (!rb) // false AND ? => false
				throw Exception(E_INVALID_SUBTYPE);
			if ( !(rb->getVal()) ) {
				delete rc; delete l;
				return new Xconstbool(false);
			}
			break;
				
		case MUL:
			if (!rr) 
				throw Exception(E_INVALID_SUBTYPE);
			if ( rr->getVal()==0.0) {
				delete rc; delete l;
				return new Xconstreal(0.0);
			}
			if ( rr->getVal()==1.0) {
				delete rc;
				return l;
			}
			break;
		
		case DIV:
			if (!rr) 
				throw Exception(E_INVALID_SUBTYPE);
			if ( rr->getVal()==0.0) {
				throw Exception(E_DIV_BY_ZERO);
			}
			if ( rr->getVal()==1.0) {
				delete rc;
				return l;
			}
			break;

		case SUB:
			if (!rr)
				throw Exception(E_INVALID_SUBTYPE);
			if ( rr->getVal()==0.0) {
				delete rc; 
				return l;
			}

		case ADD:
			if (!rr)
				throw Exception(E_INVALID_SUBTYPE);
			if ( rr->getVal()==0.0) {
				delete rc; 
				return l;
			}

	}
	return NULL;
}



// -- additional methods for simplify --- end ---



// Return string representation of object
string Xbinary::toString(int pp=0) const {
	string str_left = left->toString( op );
	string str_right = right->toString( op+1 );
	string str_op;
	
	switch (op) {
		case OR:    str_op = "or";  break;
		case AND:   str_op = "and"; break;
		case LT:    str_op = "<";   break;
		case LTEQ:  str_op = "<=";  break;
		case GT:    str_op = ">";   break;
		case GTEQ:  str_op = ">=";  break;
		case EQ:	   str_op = "=";   break;
		case NOTEQ: str_op = "!=";  break;
		case ADD:   str_op = "+";   break;
		case SUB:   str_op = "-";   break;
		case MUL:   str_op = "*";   break;
		case DIV:   str_op = "/";   break;
		case POW:	str_op = "^";   break;
		default: str_op = "bin_op"; break;
	}
		
	if (pp>op) return "("+str_left+" "+str_op+" "+str_right+")";
	else return str_left+" "+str_op+" "+str_right;
}


// Clone method - uses copy-constructor
pXpr Xbinary::clone() const {
	return new Xbinary(*this);
}


// Support for comparing - op, left and right must all match
bool Xbinary::isEqual(pXpr expr) const {
	Xbinary* p = dynamic_cast<Xbinary*>(expr);
	if (p) return (p->op == op && left->isEqual(p->left) && right->isEqual(p->right) );
	else return false;
}


// Substituting expressions - substitute in internal vars
pXpr Xbinary::subst( Xvar& x, pXpr p) const {
	return new Xbinary( left->subst(x,p), op, right->subst(x,p) );
}




///////////////////////////////////////////////////////////////////////////////
//
//  X i f t h e n e l s e - tertiary operators
//
///////////////////////////////////////////////////////////////////////////////


// This constructor will not do deep copying
Xifthenelse::Xifthenelse( pXpr ccond, pXpr oontrue, pXpr oonfalse): cond(ccond), ontrue(oontrue), onfalse(oonfalse) {
	assert( cond && ontrue && onfalse );
}


// Deep-copying constructor
Xifthenelse::Xifthenelse( const Xifthenelse& b ) {
	cond = (b.cond)->clone();
	ontrue = (b.ontrue)->clone();	
	onfalse = (b.onfalse)->clone();
	assert( cond && ontrue && onfalse );
}


// Destructor - delete the three sub-trees
Xifthenelse::~Xifthenelse() { 
	DESTRUCTOR("~Xifthenelse: " << this );
	if (cond) delete cond;
	if (ontrue) delete ontrue;
	if (onfalse) delete onfalse;
};


// Differentiates according to simple rules. Throws exception if opType is OR/AND/LT/LTEQ/GT/GTEQ/NOTEQ
pXpr Xifthenelse::diff( Xvar& x ) const {
	throw Exception(E_DIFF_IFTHENELSE);
}


// Attempts to evaluate expression. Exception thrown if condition is not of type bool!
Value Xifthenelse::evalf( Enviroment& env ) const {
	Value e = cond->evalf( env );
	if (e.type!=Value::BOOL)
		throw Exception(E_IFCOND_NOT_BOOL);
	
	if (e.val.bval)
		return ontrue->evalf( env );
	else
		return onfalse->evalf( env );
}


// get type for this expression. If type-error exists (cond not of type BOOL), NONE is returned
Xpr::exprType Xifthenelse::getType() const {
	if (cond->getType() != BOOL)
		return NONE;
	else {
		exprType et = ontrue->getType();
		exprType ef = onfalse->getType();
		if (et==ef) return et;
		else return BOTH;
	}
}


// simplifies sub-expressions and then if condition is constant then choose
pXpr Xifthenelse::simplify() const {
	DEBUG("Xifthenelse::simplify():");
	
	pXpr condsimp = NULL;
	pXpr truesimp = NULL;
	pXpr falsesimp = NULL;
	try {
		condsimp = cond->simplify();
		Xconstbool* c = dynamic_cast<Xconstbool*>(condsimp);
	
		if (c) {
			DEBUG("\tcondition fixed to value!");
			bool cv = c->getVal();
			if (cv) truesimp = ontrue->simplify();
			else truesimp = onfalse->simplify();
			delete condsimp;
			return truesimp;
		}
		
		truesimp = ontrue->simplify();
		falsesimp = onfalse->simplify();
		
		DEBUG("\tcondition not fixed, simplifying all subexpr");
  		return new Xifthenelse( condsimp, truesimp, falsesimp );
  	} catch (Exception e) {
  		if (condsimp) delete condsimp;
  		if (truesimp) delete truesimp;
  		if (falsesimp) delete falsesimp;
  		throw;
  	}
}


// Return string representation of object
string Xifthenelse::toString(int pp=0) const {
	
	string str_cond = cond->toString( 10 );
	string str_ontrue = ontrue->toString( 10 );
	string str_onfalse = onfalse->toString( 10 );
	
	if (pp>=12) return "( if "+str_cond+" then "+str_ontrue+" else "+str_onfalse+" )";
	else return "if "+str_cond+" then "+str_ontrue+" else "+str_onfalse;
}


// Clone method
pXpr Xifthenelse::clone() const {
	return new Xifthenelse(*this);
}


// Support for comparing - all three subtrees must match
bool Xifthenelse::isEqual(pXpr expr) const {
	Xifthenelse* p = dynamic_cast<Xifthenelse*>(expr);
	if (p) return (cond->isEqual(p->cond) && ontrue->isEqual(p->ontrue) && onfalse->isEqual(p->onfalse) );
	else return false;
}


// Substituting expressions - substitute in the subtrees
pXpr Xifthenelse::subst( Xvar& x, pXpr p) const {
	return new Xifthenelse( cond->subst(x,p), ontrue->subst(x,p), onfalse->subst(x,p) );
}
