//////////////////////////////////////////////////////////////////////////////
//
// File: expr.h
// 
// Defines the base and extended classes for the expression trees implemented in ExpressionWorkSHOP 2000 - 
// aka. AMANDA 2010 (Algebraic MANipulation and Data Abstraction)
//
// Notice the attempt to approach Java by including clone(), toString(int=0) and isEqual()
//
#ifndef EXPR_H
#define EXPR_H

#include <string>
#include <iostream>
#include <strstream>
#include <map>
#include "errors.h"

using namespace std;

// Declaration of classes for usage in Xpr.
class Xpr;
class Xconst;
class Xconstbool;
class Xconstreal;
class Xvar;
class Xunary;
class Xbinary;
class Xifthenelse;

// Define pointer to expression (often used so typedef issued)
typedef Xpr* pXpr;


// Abstraction from the actual enviroment
typedef map<Xvar, pXpr, less<Xvar> > Enviroment;


// All calculations are performed on this type - the largest floating-point value available
typedef long double extended;


// Wrapper structure for tagged return-value
// type-field indicates which way the val-field should be interpreted - using the fval or the bval
// field in the union.
// Usage: if (value.type==BOOL) cout << value.val.bval; else value.val.fval;
struct Value {
	enum {NONE, BOOL, REAL} type;
	union {
		extended fval;
		bool bval;
	} val;
};


// Special function for indenting (used by dumpAddr())
void indenter(int);

// Declaration of Xpr
class Xpr {
	public:
		enum exprType { NONE, BOTH, BOOL, REAL };
		
		Xpr() {}
		virtual ~Xpr() {};
	
		// Return deriviate of current expression with respect to v
		virtual pXpr diff( Xvar& v ) const = 0;
		
		// Return current expression evaluated with enviroment supplied in env
		virtual Value evalf( Enviroment& ) const = 0;

		// Return simplified version of current expression
		virtual pXpr simplify() const = 0;
		
		// Return substituted version of this object. The subs-expr _will_ be copied for each
		// needed instance, so delete your own copy of the tree afterwards
		virtual pXpr subst( Xvar&, pXpr ) const = 0;

		// Return type of subexpression. Used for type-checking of trees
		virtual exprType getType() const { return NONE; }

		// Easier use of "<<". The integer passed as parameter indicates the level of priority for the operator
		// Due to the fact that the << is an operator, it cannot be declared virtual, so instead a toString()-
		// method has been introduced.
		virtual string toString(int=0) const { return "Xpr - abstract object!"; }
		friend ostream& operator<<(ostream&, Xpr&);

		// Deep-copying function
		// This method returns a pointer to a newly allocated object equal to the original
		// The reason for this method is the fact that constructors cannot be virtual, thus
		// another approach is required when copying objects whose type is unknown at compile-
		// time.
		virtual pXpr clone() const = 0;
		// Override = operator to use clone
		// friend pXpr& operator=(pXpr&, pXpr&);
		
		// Implement this function for better support for comparing.
		// The == operator cannot be virtual in a proper way, so it is 
		// implemented using a virtual function isEqual(), that compares
		// a pointer to an object with the current object using dynamic
		// typecasting with type-check (RTTI).
		bool operator==( Xpr& x) { return isEqual(&x); }		
		virtual bool isEqual(pXpr o) const { return false; }
		
		// Debug-function. Dumps address current object plus all sub-objects
		virtual void dumpAddr(int indent=0) { indenter(indent); cout << this << endl; };
		
	protected:
	private:
};




// super-class for all constants
class Xconst: public Xpr {
	public:
		// Return deriviate of current expression with respect to v
		virtual pXpr diff( Xvar& ) const=0; 
		
		// Return current expression evaluated with enviroment supplied in env
		virtual Value evalf( Enviroment& ) const=0;
		// Constants can do an evalf() without an enviroment
		virtual Value evalf() const=0;
		
		// Return simplified version of current expression
		virtual pXpr simplify() const=0;

		// Overridden clone-method
		virtual pXpr clone() const=0;

		// Return substituted version of this object. The subs-expr _will_ be copied for each
		// needed instance, so delete your own copy of the tree afterwards
		virtual pXpr subst( Xvar&, pXpr ) const =0; // { return clone(); }



};




// class implementing boolean constant, values true/false
class Xconstbool: public Xconst {
	public:
		Xconstbool(bool val): Xconst(), value(val) {};
		~Xconstbool();

		// Raise exception
		virtual  pXpr diff( Xvar& ) const;

		// Return value
		virtual  Value evalf( Enviroment& ) const;
		virtual Value evalf() const { Value e; e.type = Value::BOOL; e.val.bval = value; return e; }

		// If return itself
		virtual  pXpr simplify() const;

		// Return substituted version of this object. The subs-expr _will_ be copied for each
		// needed instance, so delete your own copy of the tree afterwards
		virtual pXpr subst( Xvar&, pXpr ) const;

		// get type for this expression - always BOOL
		virtual  exprType getType() const { return BOOL; }

		// Overridden clone-method
		virtual pXpr clone() const;
		
		// Overridden toString method
		virtual string toString(int=0) const;
		
		// easier use that evalf
		virtual bool getVal() { return value; }
		
		// support for comparing
		virtual bool isEqual(pXpr) const;

	protected:
		bool value;
};




// class implementing real constant.
class Xconstreal: public Xconst {
	public:
		Xconstreal(extended val): Xconst(), value(val) {};
		~Xconstreal();

		// Return Xconstreal(0.0)
		virtual pXpr diff( Xvar& x ) const; 

		// Return value
		virtual Value evalf( Enviroment& ) const;
		virtual Value evalf() const { Value e; e.type = Value::REAL; e.val.fval = value; return e; }

		// If return itself
		virtual pXpr simplify() const;

		// Return substituted version of this object. The subs-expr _will_ be copied for each
		// needed instance, so delete your own copy of the tree afterwards
		virtual pXpr subst( Xvar&, pXpr ) const;

		// get type for this expression - always REAL
		virtual exprType getType() const { return REAL; }

		// Overridden clone-method
		virtual pXpr clone() const;
		
		// Overridden toString() method
		virtual string toString(int=0) const;
		
		// easier use than evalf
		extended getVal() { return value; }

		// support for comparing
		virtual bool isEqual(pXpr) const;

	protected:
		extended value;
};




// An Xvar is the basic item in an expression representing the simplest atom in the expression
class Xvar: public Xpr {
	public:
		// Construct new Xvar expression, usual copy-constructor can be used - no special processing nescesary
		Xvar(string, exprType=BOTH);
		// Copy constructor
		Xvar(const Xvar&);
		// Destroy Xvar
		~Xvar();
		
		// Differentiate current var with respect to Xvar supplied
		virtual pXpr diff( Xvar& ) const;
		
		// Evaluate this variable with respect to the enviroment. If the enviroment does not
		// contain the identifier, cause an exception since none is returned
		virtual Value evalf( Enviroment& ) const;
		
		// simplify() returns a new Xvar with same contents.
		virtual pXpr simplify() const;
		
		// Return substituted version of this object. The subs-expr _will_ be copied for each
		// needed instance, so delete your own copy of the tree afterwards
		virtual pXpr subst( Xvar&, pXpr ) const;

		// Return both, since an identifier can be both a value and a boolean
		virtual exprType getType() const { return vartype; };
		
		// ParentPriority is irrelevant
		virtual string toString(int=0) const;
		
		// Overridden clone-method
		virtual pXpr clone() const;

		// support for comparing
		virtual bool isEqual(pXpr) const;

		// support for comparing just two Xvar's
		bool operator<(const Xvar&) const;

	protected:
		string name;
		exprType vartype;
};




// Support for unary operators (a.k.a. functions like sin/cos/exp/ln/log/ee) 
class Xunary: public Xpr {
	public:
		// Normal constructor. Supply name and argument expression. Argument expression WILL NOT BE CLONED!!!
		Xunary( string, pXpr );
		// Copy constructor for deep copying. Original left and right will be cloned.
		Xunary( const Xunary& );
		// Destructor - deep delete
		~Xunary();
		
		// Differentiates according to simple rules. If it knows the function (like ln/exp/sin/cos) it differentiates accordingly, otherwise
		// it differentiates by applying ' ( like f(2*x) = 2*f'(2*x) )
		virtual pXpr diff( Xvar& ) const;

		// Attempts to evaluate expression. Known functions are calculated, NONE returned otherwise
		virtual Value evalf( Enviroment& ) const;

		// If simplify(subexpr) is Xconst and function is known then return calculated value as Xconst
		virtual pXpr simplify() const;
		
		// Return substituted version of this object. The subs-expr _will_ be copied for each
		// needed instance, so delete your own copy of the tree afterwards
		virtual pXpr subst( Xvar&, pXpr ) const;

		// get type for this expression 
		virtual exprType getType() const; 

		// Return string representation of object
		virtual string toString(int=0) const;

		// Overridden clone-method
		virtual pXpr clone() const;
		
		// support for comparing
		virtual bool isEqual(pXpr) const;

		// DEBUG - Dump addresses!
		virtual void dumpAddr(int indent=0) { 
			cout << this << "-( " << endl; 
			arg->dumpAddr(); 
			cout << ") " << endl; 
		};


	protected:
		string fname;
		pXpr arg;
		enum { OTHER, SIN, COS, EXP, LN, TAN, NOT, SINH, COSH } func;
		
		virtual Value calculate( Value v ) const;
		
};




// Implement binary operators like '+' / '-' / '*' / '/'
class Xbinary: public Xpr {
	public:
		// Notice, these enumerations may NOT be altered, they are used to determin priority in a call to toString()
		// The highest priority used by this is then 0..11
		enum opType { OR, AND, LT, LTEQ, GT, GTEQ, EQ, NOTEQ, ADD, SUB, MUL, DIV, POW };
		
		// The constructor will NOT simplify "Xconst op Xconst". Arguments WILL NOT BE COPIED!
		Xbinary( pXpr, opType, pXpr );
		// Due to use of pointer, a deep-copy is required and a new copy-constructor is created
		Xbinary( const Xbinary& );
		// Deletes both nodes
		virtual ~Xbinary();
		
		// Differentiates according to simple rules. Throws exception if opType is OR/AND/LT/LTEQ/GT/GTEQ/NOTEQ
		virtual pXpr diff( Xvar& ) const;
		
		// Attempts to evaluate expression
		virtual Value evalf( Enviroment& ) const;
		
		// simplifies sub-expressions and then "Xconst op Xconst"
		virtual pXpr simplify() const;

		// Return substituted version of this object. The subs-expr _will_ be copied for each
		// needed instance, so delete your own copy of the tree afterwards
		virtual pXpr subst( Xvar&, pXpr ) const;

		// get type for this expression. If type-error exists, NONE is returned / exception is thrown
		virtual exprType getType() const;
		
		// Return string representation of object
		virtual string toString(int=0) const;
		
		// Overridden clone-method
		virtual pXpr clone() const;

		// support for comparing
		virtual bool isEqual(pXpr) const;


		// DEBUG - Dump addresses!
		virtual void dumpAddr(int indent=0) { 
			indenter(indent); cout << this << "-( " << endl; 
			left->dumpAddr(indent+1);
			right->dumpAddr(indent+1); 
			indenter(indent); cout << ") " << endl; 
		}

	protected:
		opType op;
		pXpr left;
		pXpr right;
		
	private:
		// The passed pointers will _always_ be freed by the calling method (simplify())
		// The following three methods may throw exceptions, but are themselves responsible for cleaning up what they
		// create on the heap. 
		pXpr simp_const_op_expr(Xconst* lc, Xpr* r) const; // returns NULL if no match
		pXpr simp_expr_op_const(Xpr* l, Xconst *rc) const; // returns NULL if no match
		
		virtual Value calculate(Value, Value ) const;
};




// Implement the if-then-else expression.
class Xifthenelse: public Xpr {
	public:
		// The constructor will attempt a type-check of condition. It will NOT perform DEEP COPYING
		Xifthenelse( pXpr, pXpr, pXpr );
		// Due to use of pointer, a deep-copy is required and a new copy-constructor is created
		Xifthenelse( const Xifthenelse& );
		// Deletes all three nodes
		virtual ~Xifthenelse();
		
		// Cause exception
		virtual pXpr diff( Xvar& ) const;
		
		// Attempts to evaluate expression
		virtual Value evalf( Enviroment& ) const;
		
		// simplifies sub-expressions and if the statement can be resolved, it is resolved.
		virtual pXpr simplify() const;
		
		// Return substituted version of this object. The subs-expr _will_ be copied for each
		// needed instance, so delete your own copy of the tree afterwards
		virtual pXpr subst( Xvar&, pXpr ) const;

		// get type for this expression. If type-error exists, NONE is returned, otherwise BOTH since it cannot be determined.
		virtual exprType getType() const;
		
		// Return string representation of object
		virtual string toString(int=0) const;
		
		// Overridden clone-method
		virtual pXpr clone() const;

		// support for comparing
		virtual bool isEqual(pXpr) const;

		// DEBUG - Dump addresses!
		virtual void dumpAddr(int indent=0) { 
			indenter(indent); cout << this << "-( " << endl; 
			cond->dumpAddr(indent+1);
			ontrue->dumpAddr(indent+1);
			onfalse->dumpAddr(indent+1); 
			indenter(indent); cout << ") " << endl; 
		}

	protected:
		pXpr cond;
		pXpr ontrue;
		pXpr onfalse;
};


// Overloading the << operator for all Xpr objects
ostream& operator<<(ostream&, Xpr&);

#endif
