//////////////////////////////////////////////////////////////////////////////
//
// File: parser.cpp
//
// This file implements the parser.
// The parser is responsible for building the expression tree and issuing 
// commands on the tree. This results in the fact that Parser implements the 
// enviroment used to handle all evaluations on the expressions.
//
// The parser also includes the rules for assignment.
//
// The parser is supplied with an istream and an ostream to handle input and
// output, thus you can supply a strstream, so you may control the parser
// from another routine and use its output on-screen.

/*

Grammar for ExpressionWorkshop2000
----------------------------------


<command> ::= "quit" | "show" <E0> | "eval" <E0> | "num" <id> { <id> } | "bool" <id> { <id> } | "let" <id> "=" <metaX> | "env" | "restart" |	"help" [<command>] | <id> ":=" ["eval"] <metaX>

<metaX> ::= "diff" "(" <id> ")" <metaX> | "subst" "(" <id> [ "," <metaX> ] ")" <metaX> | "simplify" <metaX> | <E0>

<E0> ::= "if" <E0> "then" <metaX> "else" <metaX> | <E1>

<E1> ::= <E2> { "or" <E2> }

<E2> ::= <E3> { "and" <E3> }

<E3> ::= <E4> <relOp> <E4> | <E4>
<relOp> ::= "<" | "<=" | ">" | ">=" | "=" | "!="

<E4> ::= <plusminus> <E5> | <E5> { <plusminus> <E5> } | <E5>
<plusminus> = "+" | "-"

<E5> ::= <E6> <divmul> <E6> | <E6>
<divmul> ::= "*" | "/"

<E6> ::= <id> "(" <E0> ")" | "(" E0 ")" | "!" <E6> | <id> | <const>


*/
#include <iostream>
#include "scanner.h"
#include "expr.h"
#include "parser.h"
#include "errors.h"

//#define DEBUG(s) (cerr << s << endl)
#define DEBUG(s)

using namespace std;


// Call scanner and read next token
void Parser::nextTok() { 
	curTok = Sc.nexttoken(); 
}


// Compare the token t to the current token. If they are different, throw 
// an exception, otherwise read next token
void Parser::tokenP(Token t) {	
	if(curTok==t){
	nextTok();
	}
	else throw Exception(E_PS_MISSINGTOK, Sc.tokenAsString(t));
}


// This method parses a command expression of the format
// <command> ::= "quit" | "show" <E0> | "eval" <E0> | 
//							 "num" <id> { <id> } | "bool" <id> { <id> } |
//							 "let" <id> "=" <metaX> | "env" | "restart" |
//							 "help" [<command>] | <id> ":=" <metaX>
// The method will let an exception fall through and be handled outside
bool Parser::commandP() {
  	char buf[500];
	in.getline(buf,500);
	Sc.initbuf(string(buf)); 
	nextTok();
 
	switch (curTok) {
		// empty command
		case EOFI:
			break;
		
		// Quit expression workshop
		case QUIT: 
			nextTok(); 
			return false; 
			break;

		// Print enviroment
		case ENVIROMENT: {
			Enviroment::iterator p;
			for ( p=env.begin(); p!=env.end(); p++)
				if (p->second!=NULL) 
					out << p->first.toString() << " -> " << p->second->toString() << endl;
				else
					out << p->first.toString() << lang_get(EX_UNBOUNDVAR) << endl;
				break;
			}

		// Help function - uses both with or without argument
		case HELP:
			nextTok();
			switch (curTok) {
				case NUM: out << lang_get(H_NUM) << endl;break;	
				case BOOL: out << lang_get(H_BOOL) << endl;break;	
				case EVAL: out << lang_get(H_EVAL) << endl;break;	
				case LET: out << lang_get(H_LET) << endl;break;	
				case DEFEQ: out << lang_get(H_DEFEQ) << endl;break;	
				case SIMPLIFY: out << lang_get(H_SIMPLIFY) << endl;break;	
				case DIFF: out << lang_get(H_DIFF) << endl;break;
				case SUBST: out << lang_get(H_SUBST) << endl;break;
				case HELP: out << lang_get(H_HELP) << endl;break;	
				case SHOW: out << lang_get(H_SHOW) << endl;break;	
				case ENVIROMENT: out << lang_get(H_ENV) << endl;break;	
				case RESTART: out << lang_get(H_RESTART) << endl;break;	
				case QUIT: out << lang_get(H_QUIT) << endl;break;	
				case IF: case THEN:	case ELSE: 
					out << lang_get(H_IFTHENELSE) << endl;break;
				case NOT: out << lang_get(H_NOT) << endl;break;
				case EOFI: out << lang_get(H_EOF) << endl;break;
				default: out << lang_get(H_INSULT) << endl;
			}
			break;	


		// Show expression
		case SHOW: {
	 		nextTok();
			pXpr ep = prsE0(); 
			if (dynamic_cast<Xvar*>(ep)) {
				Xvar* v = dynamic_cast<Xvar*>(ep);
				Enviroment::iterator p = env.find(*v);
				if (p!=env.end()) out << p->first.toString() << " -> " << (p->second?p->second->toString():"null") << endl;
				else out << lang_get(23) << endl;
			} else {
				out << *ep << endl;
			}
			delete ep;
			break;
		}

		// reset enviroment - free memory and reset map
		case RESTART: {
			Enviroment::iterator p;
			for (p=env.begin(); p!=env.end(); p++)
				delete p->second;
			env.clear();
			out << lang_get(TX_ENVRESET) << endl;
			break;
		}

		// define number
		case NUM:
			nextTok();
			do {
				if (curTok!=ID) throw Exception(E_PS_MISSINGTOK,lang_get(TX_IDENT));
				string id = Sc.getLexeme();
				Xvar x(id, Xpr::REAL);
				if (env[x]) throw Exception(E_PS_REDEF, id);
				env[x] = new Xconstreal(0.0);
				nextTok();
			} while (curTok==ID);
			break;

		// define bool
		case BOOL:
			nextTok();
			do {
				if (curTok!=ID) throw Exception(E_PS_MISSINGTOK,lang_get(TX_IDENT));
				string id = Sc.getLexeme();
				Xvar x(id,Xpr::BOOL);
				if (env[x]) throw Exception(E_PS_REDEF, id);
				env[x] = new Xconstbool(false);
				nextTok();
			} while (curTok==ID);
			break;

		// define meta-expression
		case LET: {
			pXpr ep = NULL;
			pXpr ev = NULL;
			nextTok();
			if (curTok!=ID) throw Exception(E_PS_MISSINGTOK,lang_get(TX_IDENT));
			Xvar x(Sc.getLexeme());
			nextTok();
			tokenP(EQ);
			ep = metaX();
			ev = env[x];
			if (dynamic_cast<Xconst*>(ev)) throw Exception(E_PS_REDEF_CONST,x.toString());
			if (ev) delete ev;
			env[x] = ep;
			out << x << " -> " << *ep << endl;
			if (ep->getType()==Xpr::NONE) out << lang_get(TX_WARNTYPES) << endl;
			break;
		}

		// evaluate expression
		case EVAL:{ 	
			nextTok(); 
			pXpr ep = prsE0();
			Value res;
			try {
				if (dynamic_cast<Xvar*>(ep)) {
					Xvar* v = static_cast<Xvar*>(ep);
					Enviroment::iterator p = env.find(*v);
					if (p!=env.end()) res = p->second->evalf(env);
					else res.type = Value::NONE;
				} else
					res = ep->evalf(env);
				switch (res.type) {
					case Value::NONE: out << lang_get(22) << endl; break;
					case Value::BOOL: out << (res.val.bval?lang_get(TX_TRUE):lang_get(TX_FALSE)) << endl; break;
					case Value::REAL: out << res.val.fval << endl; break;
				}
				delete ep;
			} catch (Exception) {
				delete ep;
				throw;
			}
			break;
		}
	
		// Handle assignment
		case ID: {
			bool eval = false;
			Xvar x(Sc.getLexeme());
			nextTok();
			if (curTok!=DEFEQ) throw Exception(E_PS_MISSINGTOK, Sc.tokenAsString(DEFEQ));
			nextTok();
			if (curTok==EVAL) {
				nextTok();
				eval = true;
			}
			pXpr ep = metaX();
			if (eval) {
				Value v;
				try {	v = ep->evalf(env); } 
				catch (Exception) { delete ep; throw; }
				delete ep;
				if (v.type == Value::BOOL) ep = new Xconstbool(v.val.bval);
				if (v.type == Value::REAL) ep = new Xconstreal(v.val.fval);
			}
			if (!ep) throw Exception(E_PS_MISSINGTOK, "metaX");
			pXpr Xp = env[x];
			if (Xp) {
				if (dynamic_cast<Xconst*>(Xp)) { // enviroment bundet til constant, kan ikke lave en metadef
					// test at det nye expression er konstant og har korrekt type 
					if (dynamic_cast<Xconst*>(ep) && ep->getType()==Xp->getType() ) {
						env[x] = ep;
						delete Xp;
					} else {
						delete ep;
						throw Exception(E_PS_REDEF_CONST, x.toString() );
					}
				} else {
					delete Xp;
					env[x] = ep;
				}
			} else 
					env[x] = ep;
			out << x << " -> " << *ep << endl;
			if (ep->getType()==Xpr::NONE) out << lang_get(TX_WARNTYPES) << endl;
			break;
		}
		
		// Other tokens undefined tokens
		default: 
	 		throw Exception(E_PS_SYNTAX, Sc.getLexeme() );
	}
	return true;
} 


// This parser parses meta-expressions such as diff(), fold() and subst()
// <metaX> ::= "diff" "(" <id> ")" <metaX> |
//             "subst" "(" <id> [ "," <metaX> ] ")" <metaX> |
//             "simplify" <metaX> | <E0>
pXpr Parser::metaX() {
	DEBUG("Parser::metaX()");
	pXpr x = NULL;
	pXpr y = NULL;
	pXpr z = NULL;
	try {
		switch (curTok) {
			case DIFF: { // syntax: "diff" "(" <id> ")" <metaX>
				nextTok();
				tokenP(LP);
				if (curTok!=ID) throw Exception(E_PS_MISSINGTOK, lang_get(TX_IDENT));
				Xvar v(Sc.getLexeme());
				nextTok();
				tokenP(RP);
				x = metaX();
				if (x) y = x->diff(v);
				else throw Exception(E_PS_MISSINGTOK, lang_get(TX_EXPR));
				delete x;
				return y;
				break;
			}
			
			case SUBST: { // syntax "subst" "(" <id> "," <metaX> ) <metaX>
				nextTok();
				tokenP(LP);
				if (curTok!=ID) throw Exception(E_PS_MISSINGTOK,lang_get(EX_MISSIDSUBST));
				Xvar v(Sc.getLexeme());
				nextTok();
				if (curTok==COMMA) {
					nextTok();
					y = metaX();
					if (!y) throw Exception(E_PS_MISSINGTOK, lang_get(EX_MISSPART));
				}
				tokenP(RP);
				x = metaX();
				if (x && !y) {
					Enviroment::iterator p;
					p = env.find(v);
					if (p!=env.end() && p->second) {
						y = p->second->clone();
					} else 
						throw Exception(E_PS_UNBOUND, v.toString() );
				}
				
				if (x && y) z = x->subst(v,y);
				else throw Exception(E_PS_UNBOUND);
				delete x;
				delete y;
				return z;
				break;
			}
			
			case SIMPLIFY: { // syntax "simplify" <metaX>
				nextTok();
				x = metaX();
				y = x->simplify();
				delete x;
				return y;
				break;
			}
							
			default:
				return prsE0();
		}
	} catch (Exception e) {
		if (x) delete x;
		if (y) delete y;
		if (z) delete z;
		throw;
	}
}
			
	
	
// Parser for E0-expressions (that is - if-then-else)
// <E0> ::= "if" <E0> "then" <metaX> "else" <metaX> | <E1>
//
// Notice the initalization of c,t,f to NULL - pointers can
// be interpreted as numbers, so if c,t or f is unassigned, 
// they will still be 0 when an exception is thrown, so only
// assigned pointers will be deleted.
pXpr Parser::prsE0() { 
	DEBUG("Parser::prsE0()");
	pXpr c = NULL;
	pXpr t = NULL;
	pXpr f = NULL;
	try {
		if (curTok == IF) {
			nextTok();
			c = prsE0(); 
			tokenP(THEN); 
			t = metaX();
			tokenP(ELSE); 
			f = metaX();
			return new Xifthenelse(c,t,f);
		 } else return prsE1();
	} catch (Exception e) {
		if (c) delete c;
		if (t) delete t;
		if (f) delete f;
		throw;
	}
}

// Parser for E1-expressions - that is OR-expressions
// <E1> ::= <E2> { "or" <E2> }
pXpr Parser::prsE1() {
	DEBUG("Parser::prsE1()");
	pXpr lop = NULL;
	pXpr rop = NULL;
	try {
		lop = prsE2();
		while (curTok==OR) {
			nextTok();
			rop = prsE2();
			lop = new Xbinary(lop, Xbinary::OR, rop);
			rop = NULL;
		}
		return lop;
	} catch (Exception e) {
		if (lop) delete lop;
		if (rop) delete rop;
		throw;
	}	
}


// Parser for E2-expressions - that is AND-expressions
// <E2> ::= <E3> { "and" <E3> }
pXpr Parser::prsE2() {
	DEBUG("Parser::prsE2()");
	pXpr lop = NULL;
	pXpr rop = NULL;
	try {
		lop = prsE3();
		while (curTok==AND) {
			nextTok();
			rop = prsE3();
			lop = new Xbinary(lop, Xbinary::AND, rop);
			rop = NULL;
		}
		return lop;		
	} catch (Exception e) {
		if (lop) delete lop;
		if (rop) delete rop;
		throw;
	}	
}



// Parser for relational operators
// <relOp> ::= "<" | "<=" | ">" | ">=" | "=" | "!="
// <E3> ::= <E4> <relOp> <E4> | <E4>
pXpr Parser::prsE3() {
	DEBUG("Parser::prsE3()");
	pXpr lop = NULL;
	pXpr rop = NULL;
	try {
		lop = prsE4();
		switch(curTok) {			
		case EQ:
			 nextTok();
			 rop= prsE4();
			 return new Xbinary(lop, Xbinary::EQ, rop); 

		case NOTEQ:
			 nextTok();
			 rop= prsE4();
			 return new Xbinary(lop, Xbinary::NOTEQ, rop); 

		case GT: 
			 nextTok(); 
			 rop= prsE4(); 
			 return new Xbinary(lop, Xbinary::GT, rop);
			 
		case GTEQ:
			 nextTok();
			 rop= prsE4();
			 return new Xbinary(lop, Xbinary::GTEQ, rop);
			 
		case LT:
			 nextTok();
			 rop= prsE4();
			 return new Xbinary(lop, Xbinary::LT, rop); 
			 
		case LTEQ:
			 nextTok();
			 rop= prsE4();
			 return new Xbinary(lop, Xbinary::LTEQ, rop); 
			 break;

		default: return lop;
		}
	} catch (Exception e) {
		if (lop) delete lop;
		if (rop) delete rop;
		throw;
	}
}


// Parser for E4-expressions - "+" and "-", the operands are left associative, this means
// that whatever is read left of the operators are evaluated before what is on the rigth.
// <plusminus> = "+" | "-"
// <E4> ::= <plusminus> <E5> | <E5> { <plusminus> <E5> }| <E5>
// 
// This parser introduces support for unary application of these operators by accepting no 
// expressions before the operator and interpreting that as a constant=0.0
pXpr Parser::prsE4() {
	DEBUG("Parser::prsE4()");
	pXpr lop = NULL;
	pXpr rop = NULL;
	
	try {
		// support for unary + and -
		switch (curTok) {
		case ADD: 
			lop = new Xconstreal(0.0);
			break;
		case SUB:
			lop = new Xconstreal(0.0);
			break;
		default:
			lop = prsE5();
		}
		while (curTok==ADD || curTok==SUB) {
			switch (curTok) {
			case ADD:
				nextTok();
				rop = prsE5();
				lop = new Xbinary(lop, Xbinary::ADD, rop);
				rop = NULL;
				break;
 
			case SUB:
				nextTok();
				rop = prsE5();
				lop = new Xbinary(lop, Xbinary::SUB, rop);
				rop = NULL;
				break;
			}
		}
		return lop;
	} catch (Exception e) {
		if (lop) delete lop;
		if (rop) delete rop;
		throw;
	}
}


// Parser for E5-expressions, this is "*" and "/"
// Unlike normal compilers, the operators have same priority and is left-associative,
// compilers have higher priority on division. This has significance in serveral
// cases, where one of the evaluators may claim division-by-zero.
//
// <divmul> ::= "*" | "/"
// <E5> ::= <E56> { <divmul> <E56> }
pXpr Parser::prsE5() {
	DEBUG("Parser::prsE5()");
	pXpr lop = NULL;
	pXpr rop = NULL;
	try {
		lop = prsE56();
		while (curTok==MUL || curTok==DIV) {
			switch (curTok) {
			case MUL:
				nextTok();
				rop = prsE56();
				lop = new Xbinary(lop, Xbinary::MUL, rop);
				rop = NULL;
				break;

			case DIV:
				nextTok();
				rop = prsE56();
				lop = new Xbinary(lop, Xbinary::DIV, rop);
				rop = NULL;
				break;
				
			default:
				throw Exception(E_PS_MISSINGEXP);
			}
		}
		return lop;
	} catch (Exception e) {
		if (lop) delete lop;
		if (rop) delete rop;
		throw;
	}
}


// Parser for E56-expressions, this is "^"
// <E56> ::= <E6> { "^" <E6> }
pXpr Parser::prsE56() {
	DEBUG("Parser::prsE56()");
	pXpr lop = NULL;
	pXpr rop = NULL;
	try {
		lop = prsE6();
		while (curTok==POW) {
			nextTok();
			rop = prsE6();
			lop = new Xbinary(lop, Xbinary::POW, rop);
			rop = NULL;
			break;
		}
		return lop;
	} catch (Exception e) {
		if (lop) delete lop;
		if (rop) delete rop;
		throw;
	}
}


// Parser for E6-expressions.
// <const> ::= "true" | "false" | <number>  
// <E6> ::= <id> "(" <E0> ")" | "(" E0 ")" | "!" <E6> | <id> | <const>
pXpr Parser::prsE6() {
	DEBUG("Parser::prsE6() - " << Sc.tokenAsString(curTok) << " " << Sc.getLexeme());
	pXpr ep = NULL;
	try {
		switch (curTok) {
		case NOT:
			nextTok();
			ep = prsE6();
			return new Xunary("!",ep);
			
		case LP:
			nextTok(); 
			ep= prsE0(); 
			tokenP(RP); 
			return ep;
			 
		case CONST: {
			extended temp = Sc.getConstreal();
			ep= new Xconstreal(temp);
			nextTok();
			return ep;
		}
		
		case ID: { // <id> "(" <E0> ")" | <id>
			string id= Sc.getLexeme(); 
			nextTok();
			if (curTok==LP) {
				nextTok();
				ep = prsE0();
				if (curTok!=RP) throw Exception(E_PS_MISSINGTOK,Sc.tokenAsString(RP));
				nextTok();
				if (ep)
					return new Xunary(id, ep);
				else throw Exception(E_PS_MISSINGEXP, id);
			} else {
				Enviroment::iterator p;
				p = env.find(Xvar(id));
				if (p!=env.end()) {
					if (!(dynamic_cast<Xconst*>(p->second))) return p->second->clone();
					else return p->first.clone();
				} else throw Exception(E_PS_UNDEFID,id);
			}
			break;
		}

		case TRUE: 
				nextTok();
				return new Xconstbool(true);

		case FALSE: 
				nextTok();
				return new Xconstbool(false);

		case EOFI:
				throw Exception(E_PS_ENDOFINPUT);

		default: 
				throw Exception(E_PS_SYNTAX, Sc.getLexeme() );
		}
	} catch (Exception e) {
		if (ep) delete ep;
		throw;
	}
}
