//////////////////////////////////////////////////////////////////////////////
//
// File: scanner.cpp
//
// Implementation of a scanner for ExpressionWorkshop2000.
//
// The scanner is line-based an initialised with a line, which is then 
// scanned. Unlike the original scanner, we redefined the source to be
// a string.
/*

Grammar for Scanner
-------------------
Case-insensitive, all converted to lowercase

<digit> ::= '0'..'9'
<alpha> ::= 'a'..'z'
<alnum> ::= <alpha>, <digit>

<lt>   ::= "<"
<lteq> ::= "<="
<gt>   ::= ">"
<gteq> ::= ">="
<eq>   ::= "="
<noteq>::= "!="
<defeq>::= ":="
<not>  ::= "!"

<id> ::= <alpha> { <alnum> }
<number> ::= "0" [<base-selector>] <digit_basen> { <digit_basen> } ["." { <digit_basen> } ] | 
              <digit_base10> { <digit_base10> } [ "." { <digit_base10> } ]
<base_selector> ::= 'b' | 'o' | 'd' | 'h'
<digit_basen> ::= '0'..'9', 'a'..'f'

*/
#include <iostream>
#include <string>
#include <map>
#include <ctype.h>
#include "scanner.h"
#include "errors.h"

using namespace std;

static map<string,int> keywords;


// Scanner's constructor initializes the keyword map
Scanner::Scanner() {
  keywords["num"] = NUM;
  keywords["show"] = SHOW;
  keywords["eval"] = EVAL;
  keywords["let"] = LET;
  keywords["exit"] = QUIT;
  keywords["quit"] = QUIT;
  keywords["subst"] = SUBST;
  keywords["help"] = HELP;
  keywords["bool"] = BOOL;
  keywords["if"] = IF;
  keywords["then"] = THEN;
  keywords["else"] = ELSE;
  keywords["or"] = OR;
  keywords["and"] = AND;
  keywords["not"] = NOT;
  keywords["true"] = TRUE;
  keywords["false"] = FALSE;
  keywords["simplify"] = SIMPLIFY;
  keywords["env"] = ENVIROMENT;
  keywords["restart"] = RESTART;
  keywords["diff"] = DIFF;
  
}


// If the current token is an identifier, it is compared to the list of keywords to translate 
// it into a keyword-token instead.
void Scanner::check_keyword() { 
  map<string,int>::iterator p = keywords.find(lexeme);
  if (p!=keywords.end())
    tok = Token(p->second);
}


// This method initializes the buffer used to tokenize the input-string.
// All characters are translated to lowercase, to make the program case-insensitive.
void Scanner::initbuf( string s ) {
	linepos = 0;
	for (string::iterator i=s.begin(); i!=s.end(); i++, linepos++)
		linebuf[linepos] = tolower(*i);
	linebuf[linepos]='\0';
	linepos=0;
	nextchar();
}


// Read the next character from the input-buffer.
void Scanner::nextchar(){ 
	lexeme+= curchar; 
	curchar= linebuf[linepos++]; 
}


// Return the currently read lexeme
string Scanner::getLexeme() { 
	return lexeme;
}


// If the currently read token is a real constant return its value here
extended Scanner::getConstreal() { 
	return constreal;
}


// Read and return the type of the next token
Token Scanner::nexttoken() {
	scLexeme(); 
	return tok;
}


// Read a lexeme
// Rules are:
//  - skip whitespaces
//  - if current char is digit read a constant
//  - if current char is alphabetic read identifier
//  - if current char is special char, read appropriate function
void Scanner::scLexeme() { 
	// skip whitespaces
	while(isspace(curchar))
		nextchar(); 
		
	// reset string lexeme
	lexeme="";  
  
	// if current char is digit read constant
	if(isdigit(curchar))
		scIconst(); 
	// if current char is alphanummeric read identifier
	else if(isalpha(curchar)){
		nextchar(); 
		scId();
	}
	// evaluate special characters.
	else switch (curchar) {	
		case '+': nextchar(); tok= ADD;   return;
		case '-': nextchar(); tok= SUB;   return;
		case '*': nextchar(); tok= MUL;   return;
		case '/': nextchar(); tok= DIV;   return;
		case '(': nextchar(); tok= LP;    return;
		case ')': nextchar(); tok= RP;    return;
		case '>': nextchar(); scGT(); 	break;
		case '<': nextchar(); scLT(); 	break;
		case '!': nextchar(); scNOT();    break;
		case '=': nextchar(); tok= EQ;    break;
		case ':': nextchar(); scEQ();		break;
		case ',': nextchar(); tok= COMMA; return;
		case '^': nextchar(); tok= POW; return;
		case 0  :             tok= EOFI;  return;
		default:
			tok = ERROR;
			nextchar(); 
			throw Exception(E_SC_INVCHAR);
		}
}


// Scan an identifier into lexeme. Check afterwards if it is a keyword. (case insensitive)
// <alpha> ::= "a".."z" 
// <alphanum> ::= "0".."9", "a".."z"
// <id> ::= <alpha> { <alphanum> }
void Scanner::scId() {
	while(isalnum(curchar)){nextchar();} 
	tok = ID; 
	check_keyword();
}


// Helper-function for reading number.
bool isdelim(char c) {
  return !isalnum(c);
}


// Special function for reading constant. Supports different base-numbers. If
// a "digit" with a value larger than or equal to base is supplied, an 
// exception is thrown.
//
// <base_selector> ::= "b" | "o" | "d" | "h"
// <digit_basen> ::= "0" .. "9" | "A" .."F"
// <digit_base10> ::= "0" .. "9"
// <number> ::= "0" [<base-selector>] <digit_basen> { <digit_basen> } ["." { <digit_basen> } ] | 
//              <digit_base10> { <digit_base10> } [ "." { <digit_base10> } ]
// There is nothing wrong with using a different base like hex for decimal values - eg.
//
// 0b0100.001 = 1*2^2 + 1*2^(-3) = 4.125
// 0hAF.0F = 10*16+15+15*16^(-2) = 
void Scanner::scIconst(){ 
	extended base = 10;
	int digval = 0;
	extended fract = 0.0;
	extended intpart = 0.0;
	extended expon = 0.0;
	bool decpoint = false;

	if (curchar=='0') {
		nextchar();
		switch (curchar) {
		case 'h': base=16; nextchar(); break;
		case 'b': base=2; nextchar(); break;
		case 'd': base=10; nextchar(); break;
		case 'o': base=8; nextchar(); break;
		default: base=10;
		}
	}
	expon = 1.0/base;

	while ( !isdelim(curchar) || curchar=='.') {
		if (curchar>='0' && curchar<='9') digval=curchar-'0'; 
		if (curchar>='A' && curchar<='Z') digval=curchar-'A'+10;
		if (curchar>='a' && curchar<='z') digval=curchar-'a'+10;
		if (digval>=base) throw Exception(E_SC_INVCHAR);
		if (decpoint && curchar=='.') throw Exception(E_SC_DECPOINT);
		if (curchar=='.') {
			decpoint=true;
			nextchar();
			continue;
		}
		if (!decpoint) intpart = intpart*base + digval;
		else {
			fract += digval*expon;
			expon /= base;
		}
		nextchar();
	}
	constreal = intpart+fract;
	tok = CONST;
}


// Look for the greater-than (">") or greater-than-or-equal-to (">=")
void Scanner::scGT(){ 
	if(curchar=='=') {
		nextchar(); 
		tok=GTEQ;
	}
	else{tok=GT;} 
}


// Look for the less-than ("<") or less-than-or-equal-to ("<=")
void Scanner::scLT(){ 
	if(curchar=='='){
		nextchar(); 
		tok=LTEQ;
	} else {
		tok=LT;
	} 
}


// Check that "!" is followed by an "=" to make an NOTEQ token
void Scanner::scNOT(){
	if(curchar=='=') {
		nextchar();
		tok=NOTEQ;
	}
//	else if(isalpha(curchar)) tok = NOT;
//	else throw Exception(E_SC_INVNOT);
	else tok = NOT;
}
	  

// Check for a DEFEQ (":=") token	  
void Scanner::scEQ(){
	if(curchar != '=') throw Exception(E_SC_INVDEF);
	else {  
		nextchar();
		tok=DEFEQ;
	}
}


// Helper function to translate a token value into a string
string Scanner::tokenAsString(Token tok) { 
	switch (tok) {
	case SHOW: return("show"); break;
	case EVAL: return("eval"); break;
	case NUM: return("num"); break;
	case QUIT: return("quit"); break;
	case SIMPLIFY: return("simplify"); break;
	case SUBST: return("subst"); break;
	case ENVIROMENT: return("env"); break;
	case IF: return("if"); break;
	case THEN: return("then"); break;
	case ELSE: return("else"); break;
	case AND: return("and"); break;
	case OR: return("or"); break;
	case NOT: return("not"); break;
	case GT: return(">"); break;
	case GTEQ: return(">="); break;
	case NOTEQ: return("!="); break;
	case LT: return("<"); break;
	case LTEQ: return("<="); break;
	case ADD: return("+"); break;
	case SUB: return("-"); break;
	case MUL: return("*"); break;
	case DIV: return("/"); break;
	case POW: return("^"); break;
	case LP: return("("); break;
	case RP: return(")"); break;
	case ID: return("id"); break;
	case CONST: return("const"); break;
	case TRUE: return("true"); break;
	case FALSE: return("false"); break;
	case EOFI: return("end-of-line"); break;
	case DEFEQ: return(":="); break;
	default: return("???");
	}
}

